use structopt::StructOpt;
use tokio::{
	net::TcpStream,
	sync::watch,
};

use crate::prelude::*;
use protocol::PeerId;

/// A partially-filled-out config that structopt can deal with
/// Try to turn this into a Config as soon as possible.
#[derive (Debug, StructOpt)]
pub struct Opt {
	#[structopt (long)]
	relay_addr: Option <String>,
	#[structopt (long)]
	server_id: Option <PeerId>,
	#[structopt (long)]
	debug_echo: bool,
	#[structopt (long)]
	cert_url: Option <String>,
	#[structopt (long)]
	use_udp_over_tcp: Option <bool>,
}

/// A filled-out config for constructing an end server
#[derive (Clone)]
pub (crate) struct Config {
	pub debug_echo: bool,
	pub id: String,
	pub relay_addr: SocketAddr,
	pub relay_cert: Vec <u8>,
	pub use_udp_over_tcp: bool,
}

pub async fn main (args: &[OsString], shutdown_rx: Option <watch::Receiver <bool>>) -> anyhow::Result <()> {
	trace! ("executable_end_server::main");
	let opt = Opt::from_iter (args);
	let conf = opt.into_config ().await?;
	
	let (end_server, shutdown_tx) = P4EndServer::connect (conf).await?;
	
	let run_task = {
		tokio::spawn (async move {
			end_server.run ().await?;
			Ok::<_, anyhow::Error> (())
		})
	};
	
	if let Some (mut shutdown_rx) = shutdown_rx {
		while ! *shutdown_rx.borrow () {
			shutdown_rx.changed ().await?;
		}
		trace! ("P4 end server shutting down...");
		shutdown_tx.send (true)?
	}
	
	run_task.await??;
	
	trace! ("P4 end server shut down gracefully.");
	
	Ok (())
}

impl Opt {
	/// Converts self into a Config that the server can use.
	/// Performs I/O to load the relay cert from disk or from HTTP.
	/// Fails if arguments can't be parsed or if I/O fails.
	
	pub (crate) async fn into_config (self) -> anyhow::Result <Config> {
		let id = self.server_id.clone ().unwrap_or_else (|| "bogus_server".to_string ());
		
		let relay_addr: SocketAddr = self.relay_addr.clone ().unwrap_or_else (|| String::from ("127.0.0.1:30380")).parse ()?;
		
		// Do I/O after all parsing is done.
		// We don't want to waste a network request only to come back and error
		// out on like "127.oooo.1" not parsing into a relay address.
		
		let relay_cert: Vec <u8> = match self.cert_url.as_ref () {
			Some (url) => reqwest::get (url).await?.bytes ().await?.into_iter ().collect (),
			None => tokio::fs::read ("ptth_quic_output/quic_server.crt").await?,
		};
		
		Ok (Config {
			debug_echo: self.debug_echo,
			id,
			relay_addr,
			relay_cert,
			use_udp_over_tcp: self.use_udp_over_tcp.unwrap_or (false),
		})
	}
}

pub struct P4EndServer {
	conf: Config,
	conn: quinn::Connection,
	endpoint: quinn::Endpoint,
	shutdown_rx: watch::Receiver <bool>,
}

impl P4EndServer {
	pub (crate) async fn connect (conf: Config) -> anyhow::Result <(Self, watch::Sender <bool>)> {
		debug! ("P4 end server making its QUIC endpoint");
		let endpoint = make_client_endpoint ("0.0.0.0:0".parse ()?, &[&conf.relay_cert])?;
		
		let conf = if conf.use_udp_over_tcp {
			let udp_sock = UdpSocket::bind (SocketAddrV4::new (Ipv4Addr::UNSPECIFIED, 0)).await?;
			udp_sock.connect ((Ipv4Addr::LOCALHOST, endpoint.local_addr ()?.port ())).await?;
			
			let udp_local_server_port = udp_sock.local_addr ()?.port ();
			
			let tcp_sock = TcpSocket::new_v4 ()?;
			let tcp_conn = tcp_sock.connect (conf.relay_addr).await?;
			
			tokio::spawn (async move {
				udp_over_tcp::client::main_with_sockets (udp_sock, tcp_conn).await
			});
			
			Config {
				debug_echo: conf.debug_echo,
				id: conf.id,
				relay_addr: SocketAddr::V4 (SocketAddrV4::new (Ipv4Addr::LOCALHOST, udp_local_server_port)),
				relay_cert: conf.relay_cert,
				use_udp_over_tcp: true,
			}
		}
		else {
			conf
		};
		
		debug! ("P4 end server connecting to P3 relay server");
		let conn = protocol::p4_connect_to_p3 (
			&endpoint, 
			conf.relay_addr, 
			&conf.id
		).await?;
		
		debug! ("Connected to relay server");
		
		let (shutdown_tx, shutdown_rx) = watch::channel (false);
		
		Ok ((P4EndServer {
			conf,
			conn,
			endpoint,
			shutdown_rx,
		}, shutdown_tx))
	}
	
	pub (crate) async fn run (self) -> anyhow::Result <()> {
		trace! ("Accepting bi streams from P3");
		
		let mut shutdown_rx = self.shutdown_rx.clone ();
		
		let conf = Arc::new (self.conf);
		
		loop {
			tokio::select! {
				_ = shutdown_rx.changed () => {
					if *shutdown_rx.borrow () {
						trace! ("P4 incoming bi streams task caught graceful shutdown");
						break;
					}
				}
				stream_opt = self.conn.accept_bi () => {
					let (relay_send, relay_recv) = stream_opt?;
					
					tokio::spawn (handle_bi_stream (Arc::clone (&conf), relay_send, relay_recv));
				}
			};
		}
		
		Ok (())
	}
	
	pub fn shutting_down (&self) -> bool {
		*self.shutdown_rx.borrow ()
	}
}

async fn handle_bi_stream (
	conf: Arc <Config>,
	relay_send: quinn::SendStream,
	mut relay_recv: quinn::RecvStream,
) -> anyhow::Result <()>
{
	match protocol::p4_accept_p3_stream (&mut relay_recv).await? {
		protocol::P3ToP4Stream::NewPtthConnection {
			client_id,
			..
		} => handle_new_ptth_connection (conf, relay_send, relay_recv, client_id).await?,
	}
	
	Ok (())
}

async fn handle_new_ptth_connection (
	conf: Arc <Config>,
	mut relay_send: quinn::SendStream,
	mut relay_recv: quinn::RecvStream,
	_client_id: String,
) -> anyhow::Result <()>
{
	// TODO: Check authorization for P2 --> P4
	
	protocol::p4_authorize_p2_connection (&mut relay_send).await?;
	let p4_to_p5_req = protocol::p4_expect_p5_request (&mut relay_recv).await?;
	
	// TODO: Check authorization for P1 --> P5
	
	protocol::p4_authorize_p1_connection (&mut relay_send).await?;
	
	debug! ("Started PTTH connection");
	
	if conf.debug_echo {
		relay_send.write (b"Connected to P4=P5 debug echo server\n").await?;
		debug! ("Relaying bytes using internal debug echo server (P4=P5)");
		tokio::io::copy (&mut relay_recv, &mut relay_send).await?;
	}
	else {
		let stream = TcpStream::connect (("127.0.0.1", p4_to_p5_req.port)).await?;
		let (local_recv, local_send) = stream.into_split ();
		
		trace! ("Relaying bytes...");
		
		let ptth_conn = crate::connection::NewConnection {
			local_send,
			local_recv,
			relay_send,
			relay_recv,
		}.build ();
		
		ptth_conn.wait_for_close ().await?;
	}
	
	Ok (())
}
